// this part was designed purely for myself
// if this is useful to you great, otherwise please don't read my bad code
// i probably should be using pathbuf or something similar but i'm lazy :p
use regex::Regex;
use std::fs::{create_dir_all, rename};
use std::path::PathBuf;

const DRY_RUN: bool = false;
// minimal patterns to quickly identify if a torrent file is for music or not
const MUSIC_PATS: &'static [&'static str] = &["FLAC Lossless", "MP3 V0", "CD Log"];

#[allow(dead_code)]
fn main() {
    sort_and_move();
}

pub fn sort_and_move() {
    let paths = std::fs::read_dir(".")
        .unwrap()
        .filter_map(Result::ok)
        .map(|dir_entry| dir_entry.path())
        .filter(|f| f.extension().unwrap_or_default() == "torrent")
        .collect::<Vec<_>>();
    if paths.len() == 0 {
        println!("no files to sort");
        return;
    }
    for path in paths {
        match categorize(path) {
            Ok(_) => (),
            Err(e) => eprintln!("{}", e),
        }
    }
}

fn categorize(path: PathBuf) -> Result<(), String> {
    categorize_anime(path).or_else(categorize_music)
}

// anime and music have to be handled separately because their formats are very different
// by returning the PathBuf in the Err we can chain the functions to handle the failure
fn categorize_anime(path: PathBuf) -> Result<(), PathBuf> {
    // regex validation: https://regex101.com/r/qp0xQV/1
    let re = Regex::new(
        r".* - (?P<category>Visual Novel|Light Novel|Novel|TV S.*?|Movie|Manga|OVA|ONA|DVD Special|BD Special|Oneshot|Anthology|Artbook|Manhwa|Manhua|Artbook|Game|Live Action.*?|Pack)(?: \[[0-9]*\])?(?: - (?P<details>.+))?\.[0-9]+",
    ).unwrap();

    let file_name = path.file_name().unwrap().to_str().unwrap();
    if MUSIC_PATS.into_iter().any(|pat| file_name.contains(pat)) {
        return Err(path);
    }
    let captures = match re.captures(file_name).ok_or_else(|| 0) {
        Ok(val) => val,
        Err(_) => {
            return Err(path);
        }
    };
    let category = captures.name("category").map_or("", |m| m.as_str());
    let source_details = captures.name("details").map_or("", |f| f.as_str());

    let folder_path = match category {
        "Visual Novel" => "./torrents/vn/".to_string(),
        "Light Novel" | "Novel" => {
            if source_details.starts_with("Raw") {
                "./torrents/books/raw/".to_string()
            } else {
                "./torrents/books/".to_string()
            }
        }
        "TV Series" | "TV Special" | "BD Special" | "DVD Special" | "OVA" | "ONA" | "Movie" => {
            let folder_path = "./torrents/anime/".to_string();

            let folder_path = if source_details.contains("Hentai") {
                format!("{}{}", folder_path, "h/")
            } else {
                folder_path
            };

            let folder_path = if source_details.contains("Blu-ray") {
                format!("{}{}", folder_path, "bluray/")
            } else if source_details.contains("DVD") {
                format!("{}{}", folder_path, "dvd/")
            } else if source_details.starts_with("TV") {
                format!("{}{}", folder_path, "tv/")
            } else if source_details.starts_with("VHS") {
                format!("{}{}", folder_path, "vhs/")
            } else if source_details.starts_with("VCD") {
                format!("{}{}", folder_path, "vcd/")
            } else if source_details.starts_with("Web") {
                format!("{}{}", folder_path, "webdl/")
            } else if source_details.starts_with("LD") {
                format!("{}{}", folder_path, "ld/")
            } else {
                folder_path
            };

            let folder_path = if source_details.contains("Episode") {
                format!("{}{}", folder_path, "episode/")
            } else {
                folder_path
            };

            if source_details.contains("RAW") {
                format!("{}{}", folder_path, "raw/")
            } else {
                folder_path
            }
        }
        "Manga" | "Oneshot" | "Anthology" | "Manhwa" | "Manhua" => {
            let folder_path = if file_name.contains("Hentai") {
                "./torrents/manga/h/"
            } else {
                "./torrents/manga/"
            };

            if source_details.starts_with("Raw") {
                format!("{}{}", folder_path, "raw/")
            } else {
                folder_path.to_string()
            }
        }
        "Artbook" => "./torrents/artbooks/".to_string(),
        "Game" => "./torrents/games/".to_string(),
        "Pack" => "./torrents/packs/".to_string(),
        category if category.starts_with("Live Action") => {
            let folder_path = "./torrents/live_action/".to_string();

            let folder_path = if source_details.contains("Blu-ray") {
                format!("{}{}", folder_path, "bluray/")
            } else if source_details.contains("DVD") {
                format!("{}{}", folder_path, "dvd/")
            } else if source_details.starts_with("TV") {
                format!("{}{}", folder_path, "tv/")
            } else if source_details.starts_with("VHS") {
                format!("{}{}", folder_path, "vhs/")
            } else if source_details.starts_with("VCD") {
                format!("{}{}", folder_path, "vcd/")
            } else if source_details.starts_with("Web") {
                format!("{}{}", folder_path, "webdl/")
            } else if source_details.starts_with("LD") {
                format!("{}{}", folder_path, "ld/")
            } else {
                folder_path
            };

            if source_details.contains("RAW") {
                format!("{}{}", folder_path, "raw/")
            } else {
                folder_path
            }
        }
        _ => {
            eprintln!("[ERROR] unhandled category:  {}", category);
            return Ok(());
        }
    };
    move_to_dir(path.to_owned(), folder_path, file_name).unwrap();

    Ok(())
}

fn categorize_music(path: PathBuf) -> Result<(), String> {
    // regex validation: https://regex101.com/r/6xm7gP/1
    let re =
        Regex::new(r".* - .*?(?: \[[0-9]{4}\])(?: - (?P<details>.+))?\.[0-9]+\.torrent").unwrap();

    let file_name = path.file_name().unwrap().to_str().unwrap();
    let captures = match re.captures(file_name).ok_or_else(|| 0) {
        Ok(val) => val,
        Err(_) => {
            return Err(format!(
                "[ERROR] failed to capture anything from \"{}\"",
                file_name
            ))
        }
    };

    let source_details = captures.name("details").unwrap().as_str();
    let video_pats = vec![
        "Web",
        "TV",
        "Blu-ray",
        "DVD",
        "HD DVD",
        "UHD Blu-ray",
        "VHS",
        "VCD",
        "LD",
    ];
    let folder_path = if video_pats.iter().any(|f| source_details.starts_with(f)) {
        "./torrents/music/videos/".to_owned()
    } else {
        "./torrents/music/".to_owned()
    };
    move_to_dir(path.to_owned(), folder_path, file_name).unwrap();

    Ok(())
}

fn move_to_dir(path: PathBuf, output_dir_path: String, file_name: &str) -> std::io::Result<()> {
    if DRY_RUN {
        println!(
            "{} -> {}{}",
            path.to_str().unwrap(),
            output_dir_path,
            file_name
        );
        return Ok(());
    }

    create_dir_all(output_dir_path.to_owned())?;
    let new_location = format!("{}{}", output_dir_path, file_name);
    rename(path, new_location)
}
